import std.stdio;
import std.file;
import std.string;
import std.conv : to;


import gtk.Main;
import gtk.MainWindow;
import gtk.TreeView;
import gtk.ListStore;
import gtk.TreeViewColumn;
import gtk.TreeIter;
import gtk.Label;
import gtk.Entry;
import gtk.CellRenderer;
import gtk.CellRendererText;
import gtk.VBox;

import gtk.Button;





void main(string[] args)
{
	Main.init(args);

	//Window
	auto window = new MainWindow("2DA-Edit");

	auto cont = new VBox(false, 0);
	window.add(cont);

	auto buttonSave = new Button("document-save-symbolic", GtkIconSize.MENU);
	auto buttonOpen = new Button("document-open-symbolic", GtkIconSize.MENU);

	//"user-trash-symbolic"
	//"format-text-direction-ltr-symbolic"
	//"view-list-symbolic"
	version(Windows){
		//Menu bar
		import gtk.HBox;

		auto cont2 = new HBox(false, 0);
		cont.packStart(cont2, false, false, 0);

		cont2.packStart(buttonOpen, false, false, 0);
		cont2.packEnd(buttonSave, false, false, 0);
	}
	else{
		//Header bar
		import gtk.HeaderBar;

		auto header = new HeaderBar();
		window.setTitlebar(header);
		header.setTitle("2DAEdit");
		header.setSubtitle(args[1]);
		header.setProperty("show-close-button", true);

		header.packStart(buttonOpen);
		header.packEnd(buttonSave);

	}

	//TreeView to display database
	auto tree = new TreeView();
	cont.packEnd(tree, true, true, 0);

	tree.setHeadersVisible(true);
	tree.setEnableSearch(true);
	tree.setProperty("enable-grid-lines", GtkTreeViewGridLines.VERTICAL);
	tree.setProperty("tooltip-column", 0);
	tree.setProperty("reorderable", true);
	tree.setProperty("headers-clickable", true);


	//Configure button callbacks
	buttonSave.addOnClicked((Button){
		Save(tree);
	});

	buttonOpen.addOnClicked((Button){
		import gtk.Dialog;
		import gtk.FileChooserDialog;

		auto fc = new FileChooserDialog("Open 2DA", window, FileChooserAction.OPEN);
		auto res = fc.run();
		if(res==GtkResponseType.OK){
			string filename = fc.getFilename();
			Open(filename, tree);
		}
		fc.destroy();
	});

	//Open if exists
	if(exists(args[1])){
		Open(args[1], tree);
	}

	window.showAll();
	
	Main.run();

}

void Save(ref TreeView tree){
	auto store = cast(ListStore)tree.getModel();
	if(store !is null){
		TreeIter it = new TreeIter();
		store.getIterFirst(it);//TODO: fail if first is null
		do{
			for(int i=0 ; i<store.getNColumns() ; i++){
				std.stdio.write("\t",store.getValueString(it, i));
			}
			writeln();
		}while(store.iterNext(it));
	}
	else
		writeln("Nothing to save !");

}

void Open(string file, ref TreeView tree){

	auto twoda = new TwoDA(file);

	//Delete old store
	auto oldstore = cast(ListStore)tree.getModel();
	if(oldstore !is null)
		oldstore.destroy();

	//Remove columns from TreeView
	foreach(i ; 0..tree.getNColumns)
		tree.removeColumn(tree.getColumn(0));

	//Set store types
	GType type[];
	foreach(i;0..twoda.header.length)type~=GType.STRING;

	//Create new store
	auto store = new ListStore(type);
	tree.setModel(store);


	//Setup TreeView columns
	foreach(index, s ; twoda.header){

		CellRendererText cr;
		if(index==0){
			cr = new CellRendererText();
			cr.setProperty("editable", false);
			cr.setProperty("background-rgba", cast(ulong)(new GdkRGBA(0.36, 0.13, 0.4, 1.0)));
			cr.setProperty("background-set", true);
		}	
		else{
			cr = new CellRendererText();

			cr.setProperty("editable", true);

			cr.addOnEdited((string path, string newval, CellRendererText crt){
				TreeIter t = new TreeIter(tree.getModel(), path);
				store.setValue(t, cast(int)crt.getData("colnumber"), newval);
			});
		}
		cr.setData("colnumber", cast(void*)cast(int)index);


		auto col = new TreeViewColumn(s, cr, "text", cast(int)index);
		col.setResizable(true);
		//col.setReorderable(true);
		tree.appendColumn(col);
	}

	//Fill database
	TreeIter iter = new TreeIter();

	for(int i=0 ; i<=twoda.lastLine ; i++){
		store.append(iter);
		store.setValue(iter, 0, i);

		if(i in twoda.values){
			foreach(index, v ; twoda.values[i]){
				store.setValue(iter, cast(int)index+1, v);
			}
		}
		else{
			foreach(index ; 1..twoda.header.length){
				store.setValue(iter, cast(int)index+1, "_");
			}
		}
	}

	//Autosize columns
	tree.columnsAutosize();
}



class TwoDA{
	import std.regex;
	import std.file;

	this(string filepath){
		lastLine = 0;
		foreach(index, line ; readText(filepath).splitLines()){
			
			string data[]; 

			auto results = matchAll(line, rgxField);
			foreach(res ; results){
				string s;
				if(res[0][0]=='"')	data~= res[2];
				else				data~= res[1];
			}

			if(index==0){
				header = data;
				//writeln(header);
			}
			else{
				int nLine = data[0].to!int;
				values[nLine] = data[1..$];
				//writeln(values[nLine]);

				if(nLine > lastLine)
					lastLine = nLine;
			}

		}

	}


	string[] header;
	string[][uint] values;
	uint lastLine;
	enum rgxField = ctRegex!"(?:\\b([^\\s]+?)\\b|\"([^\"]+?)\")";
}

